/*****************************************************
CVAVR V2.05
Chip type               : ATmega8
AVR Core Clock frequency: 8,000000 MHz
*****************************************************/

#include <mega8.h>


#define A PORTB.6
#define B PORTB.7
#define C PORTD.5
#define D PORTD.6
#define E PORTB.0
#define F PORTB.1
#define G PORTB.2
#define DP PORTB.4
#define A1 PORTC.4
#define A2 PORTC.5
#define A3 PORTD.0
#define B1 PORTD.1
#define B2 PORTD.3
#define B3 PORTD.4

#define button0    PINC.2 // i reg
#define button1    PINC.3 // u reg

#define t1_start do{TCCR1B|=1<<CS11;}while(0)
#define t1_stop do{TCCR1B&=~(1<<CS11);}while(0)
#define t2_start do{TCCR2|=1<<CS21;}while(0)
#define t2_stop do{TCCR1B&=~(1<<CS21);}while(0)

#define scr PORTB.5
#define scr_on do{PORTB.5=1;}while(0)
#define scr_off do{PORTB.5=0;}while(0)

#define ADC_itn 16

#define shunt_res 10 //mOhms
#define amplification 10 //4LM358 
#define Ib_max_adc (256*1000/(amplification*shunt_res)) //=I_max*100, 25,6A*100

#define Rb_upr 18000 //Ohms
#define Rb_lwr 1000   //Ohms
#define Ub_max_adc ((256*(Rb_upr+Rb_lwr))/Rb_lwr) //=U_max*100, 48,64V*100


enum btn_states 
{b_prs, b_rls, b_hold} btn0, btn1;
enum dreg {setI, setU, nopr, ovc}d_mode=nopr;

bit scr_en;
unsigned char tmr[6];
unsigned int wvl,ADC0,ADC1;
unsigned int tmpU,tmpI, defU,defI, wU,wI;
unsigned char display_counter, display_buff [6]; 

eeprom unsigned int eetmp,eeU,eeI;


void ports_init ()
{
 DDRB = 0b11111111; PORTB = 0b00000000;  
 DDRC = 0b11110000; PORTC = 0b11001100;
 DDRD = 0b01111011; PORTD = 0b00000000;
}
void AC_init()
{
ACSR|=1<<ACIE|1<<ACBG;
}
void ADC_init ()
{
 ADCSRA |= 1<<ADEN|1<<ADSC|1<<ADPS2|1<<ADPS1|1<<ADPS0; // div by 128
 ADMUX |= 1<<REFS1|1<<REFS0; //2,56 bg refs 
}
void int0_init()
{
// INT0: On
// INT0 Mode: Any change
GICR|=0x40;
MCUCR=0x01;
GIFR=0x40;
}
void t1_cfg()
{
 TCCR1B=0x02;
 OCR1A=0x2000;
 TIMSK|=1<<OCIE1A;
}
void t2_cfg()
{
TCCR2|=1<<WGM21|1<<WGM20|1<<CS21; // /8, fast PWM
OCR2=0x7F;
}

void convert_data(unsigned int volts, unsigned int amps, unsigned char time);
void display_data ()
{

 (display_counter < 5)?(display_counter++):(display_counter = 0);
  A=B=C=D=E=F=G=DP=1; A1=A2=A3=B1=B2=B3=0;
 
 switch (display_buff[display_counter])
 {
  case 0: A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 1: A=1; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 2: A=0; B=0; C=1; D=0; E=0; F=1; G=0; break;
  case 3: A=0; B=0; C=0; D=0; E=1; F=1; G=0; break;
  case 4: A=1; B=0; C=0; D=1; E=1; F=0; G=0; break;
  case 5: A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 6: A=0; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 7: A=0; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 8: A=0; B=0; C=0; D=0; E=0; F=0; G=0; break;
  case 9: A=0; B=0; C=0; D=0; E=1; F=0; G=0; break;
  
  /*
  case 10: A=0; B=0; C=0; D=0; E=0; F=0; G=1; DP = 0; break;
  case 11: A=1; B=0; C=0; D=1; E=1; F=1; G=1; DP = 0; break;
  case 12: A=0; B=0; C=1; D=0; E=0; F=1; G=0; DP = 0; break;
  case 13: A=0; B=0; C=0; D=0; E=1; F=1; G=0; DP = 0; break;
  case 14: A=1; B=0; C=0; D=1; E=1; F=0; G=0; DP = 0; break;
  case 15: A=0; B=1; C=0; D=0; E=1; F=0; G=0; DP = 0; break;
  case 16: A=0; B=1; C=0; D=0; E=0; F=0; G=0; DP = 0; break;
  case 17: A=0; B=0; C=0; D=1; E=1; F=1; G=1; DP = 0; break;
  case 18: A=0; B=0; C=0; D=0; E=0; F=0; G=0; DP = 0; break;
  case 19: A=0; B=0; C=0; D=0; E=1; F=0; G=0; DP = 0; break;  
  case 20: A=1; B=1; C=1; D=0; E=0; F=0; G=1; break;     //L
  case 21: A=1; B=0; C=0; D=1; E=0; F=0; G=0; break;     //H
  case 22: A=1; B=1; C=1; D=1; E=1; F=1; G=1; break;     // 
  */
  default : A=1; B=1; C=1; D=1; E=1; F=1; G=0; break;    //-  
   }
 
 switch (display_counter)
 {
  case 0: A1 = 1;           break;
  case 1: A2 = 1; DP = 0;   break;
  case 2: if (!display_buff[2]) {A3 = 0;} else {A3 = 1;} break; 
  case 3: B1 = 1; break;
  case 4: B2 = 1; DP = 0; break;
  case 5: if (!display_buff[5]) {B3 = 0;} else {B3 = 1;} break;
 }
  
}

void scan_keys ();
void get_adc();
void ui_ctrl();
void mops();
void charging_process();

interrupt [EXT_INT0] void ext_int0_isr(void)
{
 unsigned char i;  
 unsigned long hwv_tmp=0;
 static unsigned int hwv[10];

 t1_stop; scr_off; 
 for(i=0;i<9;i++){hwv[i]=hwv[i+1];} hwv[9]=TCNT1;
 TCNT1=0;

 for(i=0;i<10;i++){hwv_tmp+=hwv[i];}
 wvl=hwv_tmp/10;
 
 for(i=0;i<6;i++){if(tmr[i]){tmr[i]--;}}  
  
 while(ADCSRA.ADSC);
 ADMUX&=0b11110000;
 ui_ctrl();
 t1_start;
 ADCSRA|=1<<ADSC;
}
interrupt [ANA_COMP] void ana_comp_isr(void)
{
 if(!ACSR.ACO){scr_off;scr_en=0;d_mode=ovc;tmr[5]=250;}
}
interrupt [TIM1_COMPA] void timer1_compa_isr(void)
{
 if(scr_en){scr_on;}
}


void main(void)
{
ports_init();
AC_init  ();
ADC_init ();
int0_init();
t1_cfg();
t2_cfg();

scr_en=0;
if(eetmp!=10){eetmp=10;}
if(eeU==0xff){eeU=136;}
if(eeI==0xff){eeI=55;}
defU=eeU; defI=eeI;

// Watchdog Timer initialization
// Watchdog Timer Prescaler: OSC/32k
#pragma optsize-
WDTCR=0x19;
WDTCR=0x09;
#ifdef _OPTIMIZE_SIZE_
#pragma optsize+
#endif

#asm("sei")

while (1)
      {
        get_adc();
        display_data ();
        scan_keys ();
        mops(); 
        
        
        #asm("wdr")
      }
}

void scan_keys ()
{
static unsigned int count0, count1, count3;

if (button0){ btn0 = b_rls;}
if (button1){ btn1 = b_rls;}

if (!button0) {count0++; } 
else {count0 = 0;}
if (!button1) {count1++; } 
else {count1 = 0;}

if ((count0 > 155)&&(btn0 != b_hold)) {btn0 = b_prs; count0 = 0; count3++;}
if ((count1 > 155)&&(btn1 != b_hold)) {btn1 = b_prs; count1 = 0; count3++;} 

if ((count3 > 25)&&(btn0 == b_prs)) {btn0 = b_hold; count3 = 0;}
if ((count3 > 25)&&(btn1 == b_prs)) {btn1 = b_hold; count3 = 0;}
}
void get_adc()
{
 static unsigned int ADC0_tmp [ADC_itn],ADC1_tmp [ADC_itn];
 unsigned long dtmp=0;
 unsigned char i;

if(!ADCSR.ADSC)
{    
  if (!(ADMUX&1))
 {
  ADMUX |= 1;  
  for (i=0;i<(ADC_itn-1);i++) {ADC0_tmp[i]=ADC0_tmp[i+1];}
  ADC0_tmp[(ADC_itn-1)] = ADCW;
  for (i=0; i<(ADC_itn); i++ ) {dtmp+=ADC0_tmp[i];}
  ADC0 = dtmp/ADC_itn; 
  ADCSR |= 1<<ADSC; 
 } 
 else
 {
  for (i=0;i<(ADC_itn-1);i++) {ADC1_tmp[i]=ADC1_tmp[i+1];}
  ADC1_tmp[(ADC_itn-1)] = ADCW;
  for (i=0; i<(ADC_itn);i++) {dtmp+=ADC1_tmp[i];}
  ADC1 = dtmp/ADC_itn; 
  ADCSR |= 1<<ADSC; 
 }
}
}
void ui_ctrl()
{
 tmpU=((unsigned long)ADC0*Ub_max_adc/(0x3ff*10));
 tmpI=((unsigned long)ADC1*Ib_max_adc/(0x3ff*10)); 
     

 if((tmpU>wU+100)||(tmpI>wI+100))
    {
    if(OCR1A<wvl-1050){OCR1A+=1000;}else if(OCR1A<wvl-130){OCR1A+=100;}
    else if(OCR1A<wvl-25){OCR1A+=10;}else if(OCR1A<wvl-20){OCR1A++;}
    }
 else if((tmpU>wU+10)||(tmpI>wI+10))
 {if(OCR1A<wvl-25){OCR1A+=10;}else if(OCR1A<wvl-20){OCR1A++;}}
 else if((tmpU>wU+1)||(tmpI>wI+1)){if(OCR1A<wvl-20){OCR1A++;}}

 
 else if((scr_en)&&(wU>50)&&(wI>50)&&(tmpU<wU-50)&&(tmpI<wI-50))
 {if(OCR1A>7500){OCR1A-=250;} else if(OCR1A>500){OCR1A-=10;}else if(OCR1A>5){OCR1A--;}}
 else if((scr_en)&&(tmpU<wU)&&(tmpI<wI))
 {if(OCR1A>5){OCR1A--;}}

}
void convert_data (unsigned int volts, unsigned int amps, unsigned char time)
{
 char i=0;

 if(!tmr[0]) 
 {
  tmr[0]=time;

    for (i=0; i<3; i++) {display_buff [i] = volts%10; volts /= 10;}
    for (i=3; i<6; i++) {display_buff [i] = amps%10; amps /= 10;}
 }
}
void mops()
{
 static bit trg0,trg1; 
 
 { 
 if((btn0==b_rls)&&(btn1==b_rls)){trg1=0;}  
 if(((btn0==b_prs)||(btn0==b_hold))&&(!trg0)&&(d_mode!=setU))
 {trg0=1;d_mode=setI;}   
 
 if((btn0==b_rls)&&(btn1==b_rls)){trg1=0;}  
 if(((btn1==b_prs)||(btn1==b_hold))&&(!trg0)&&(d_mode!=setI))
 {trg0=1;d_mode=setU;}                 
 }
 
switch (d_mode)
{
 case setI: 
 {
   if((btn0==b_prs)||(btn0==b_hold)){tmr[1]=200;scr_en=0;}
   if(tmr[1])
   {
    if((btn0==b_prs)&&(!trg1))
    {
     trg1=1; if(defI<(Ib_max_adc/10)-5){defI+=5;}else defI=5; 
     convert_data (defU,defI,0);
    } 
    if(btn0==b_hold)
    {
      if (!tmr[2])
      {
      tmr[2]=25;if(defI<(Ib_max_adc/10)-10){defI+=10;}else defI=10;
      convert_data (defU,defI,0);    
      } 
    }
   }
   else {if(defI!=eeI){eeI=defI;} d_mode=nopr; trg0=0;}
}   
 break;
 
 case setU: 
 {
   if((btn1==b_prs)||(btn1==b_hold)){tmr[3]=200;scr_en=0;}
   if(tmr[3])
   {
    if((btn1==b_prs)&&(!trg1))
    {
     trg1=1; if(defU<325){defU+=5;}else defU=10;
     convert_data (defU,defI,0);
    } 
    if(btn1==b_hold)
    {
      if (!tmr[4])
      {
      tmr[4]=25;if(defU<330){defU+=10;}else defU=10;
      convert_data (defU,defI,0);    
      }    
    }
   }
   else {if(defU!=eeU){eeU=defU;} d_mode=nopr; trg0=0;}
 }
 break;
  
 case nopr: 
 {
  convert_data (tmpU,tmpI,100);
  charging_process();
 }
 break;
 
 case ovc: 
 {
 if(!tmr[5]){d_mode=nopr;OCR1A=wvl-10;}
 }
 break;  
}
}
void charging_process()    
{
 wU=defU;wI=defI;
 
 if(tmpU<30){scr_en=0;} 
 else if(tmpU<=defU){scr_en=1;} 

}